package fmon.builder

import java.io._
import java.util.prefs.Preferences

import scala.util.Properties

import javafx.application.Application
import javafx.fxml.FXML
import javafx.fxml.FXMLLoader
import javafx.fxml.JavaFXBuilderFactory
import javafx.scene.{control => jfxsc, layout => jfxsl, Parent, Scene}
import javafx.stage.Stage

import scalafx.Includes._
import scalafx.beans.property._
import scalafx.collections.ObservableBuffer
import scalafx.scene.control._
import scalafx.scene.image.ImageView
import scalafx.scene.input.MouseEvent
import scalafx.scene.paint.Color
import scalafx.scene.layout._
import scalafx.stage.FileChooser
import FileChooser.ExtensionFilter

import fmon._
import fmon.draw.Palette
import fmon.draw.tile._
import fmon.util.Direction
import fmon.world._

import Direction._

import scala.math.{min, max}

class MapBuilder {
  @FXML var tileSelector: jfxsl.TilePane = _
  @FXML var doodadSelector: jfxsl.TilePane = _
  @FXML var gameMap: jfxsl.TilePane = _
  
  var imgSeq: IndexedSeq[ImageView] = _
  
  var tileset: Tileset = _
  var currTile: AutoTile = _
  
  var level: GameMap = _
  
  var lastFile: File = _
  
  def lastDir(): String = {
    val prefs = Preferences.userNodeForPackage(classOf[MapBuilder])
    prefs.get("dir", Properties.userDir)
  }
  
  def rememberDir(file: File): Unit = {
    val dir = if (file.isDirectory()) file.getPath else file.getParent
    if (dir != lastDir) {
      val prefs = Preferences.userNodeForPackage(classOf[MapBuilder])
      prefs.put("dir", dir)
    }
  }
  
  @FXML
  def initialize(): Unit = {
    tileset = new Tileset(new TilesetToken("Tileset", null, raw"C:\Users\dalyj\Documents\Design\Images\AutoTiles\tilea2.png", null, null, null, null, null, IndexedSeq()))
    level = new GameMap(10, 10, tileset)
    setup()
  }
  
  def setup(): Unit = {
    
    val icons = tileset.groundTiles.map(t => {
      val view = new ImageView(t.icon.croppedImage())
      view.handleEvent(MouseEvent.Any) {
        e: MouseEvent => if (e.eventType == MouseEvent.MouseClicked) {
          currTile = t
        }
      }
      view.delegate
    })
    tileSelector.children.clear()
    tileSelector.children ++= icons
    currTile = tileset.groundTiles.head
    
    val doodadFile = new File(raw"C:\Users\dalyj\Documents\Design\Images\Icons\equipment-icons-tileset.png")
    val doodadPalette = Palette.bySize(doodadFile, Config.tileSize, Config.tileSize)
    
    val doodads = doodadPalette.images.map(d => {
      val view = new ImageView(d.croppedImage())
      view.delegate
    })
    doodadSelector.children.clear()
    doodadSelector.children ++= doodads 
    
    imgSeq = for (y <- 0 until level.height; x <- 0 until level.width) yield {
      val view = new ImageView(currTile.icon.croppedImage()) // TODO Tile with adjacent
      view.handleEvent(MouseEvent.Any) {
        e: MouseEvent => if (e.eventType == MouseEvent.MouseClicked) {
          val index = point2index(x, y)
          level.tiles = level.tiles.updated(index, currTile)
          autotile(x, y)
        }
      }
      view
    }
    gameMap.children.clear()
    gameMap.prefColumns = level.width
    gameMap.prefRows = level.height
    gameMap.children ++= imgSeq.map(_.delegate)
    
    for (y <- 0 until level.height; x <- 0 until level.width) {
      smoothTile(x, y)
    }
  }
  
  def autotile(x: Int, y: Int) = {
    for (yy <- max(y - 1, 0) to min(y + 1, level.height - 1); 
        xx <- max(x - 1, 0) to min(x + 1, level.width - 1)) {
      smoothTile(xx, yy)
    }
  }
  
  def smoothTile(x: Int, y: Int) = {
    val i = point2index(x, y)
    val dirs = scala.collection.mutable.Set[Direction.Value]()
    
    val onLeft = x == 0
    val onRight = x + 1 == level.width
    val onTop = y == 0
    val onBottom = y + 1 == level.height
    
    if (onLeft || level.tiles(left(i)) == level.tiles(i)) dirs += West
    if (onRight || level.tiles(right(i)) == level.tiles(i)) dirs += East 
    if (onTop || level.tiles(up(i)) == level.tiles(i)) dirs += North
    if (onBottom || level.tiles(down(i)) == level.tiles(i)) dirs += South
    if (onLeft || onTop || level.tiles(left(up(i))) == level.tiles(i)) dirs += Northwest
    if (onLeft || onBottom || level.tiles(down(left(i))) == level.tiles(i)) dirs += Southwest
    if (onTop || onRight || level.tiles(right(up(i))) == level.tiles(i)) dirs += Northeast
    if (onBottom || onRight || level.tiles(right(down(i))) == level.tiles(i)) dirs += Southeast
    val index = point2index(x, y) 
    imgSeq(index).image = level.tiles(index).build(dirs.toSet).croppedImage()
  }
  
  def save() = {
    if (lastFile != null) {
      level saveTo lastFile
    } else {
      saveAs()
    }  
  }
  
  def saveAs() = {
    val fileChooser = new FileChooser {
      title = "Open Resource File"
      initialDirectory = new File(lastDir())
      extensionFilters ++= Seq(
        new ExtensionFilter("MAP Files", "*.map"),
        new ExtensionFilter("All Files", "*.*")
      )
    }
    
    val selectedFile = fileChooser.showSaveDialog(null)
    if (selectedFile != null) {
      rememberDir(selectedFile)
      level saveTo selectedFile
      lastFile = selectedFile
    }
  }
  
  def open() = {
    val fileChooser = new FileChooser {
      title = "Open Resource File"
      initialDirectory = new File(lastDir())
      extensionFilters ++= Seq(
        new ExtensionFilter("MAP Files", "*.map"),
        new ExtensionFilter("All Files", "*.*")
      )
    }
    
    val selectedFile = fileChooser.showOpenDialog(null)
    if (selectedFile != null) {
      rememberDir(selectedFile)
      level = GameMap loadFrom selectedFile
      tileset = level.tileset
      setup()
      lastFile = selectedFile
    }
  }
  
  def point2index(x: Int, y: Int) = level.pt2index(x, y)
  
  def left(i: Int) = i - 1
  def right(i: Int) = i + 1
  def up(i: Int) = i - level.width
  def down(i: Int) = i + level.width
}

object MapBuilder {
  class MBA extends Application {
    override def start(primaryStage: Stage): Unit = {
      val frameLoader = new FXMLLoader(getClass.getResource("MapBuilder.fxml"))
      val root: Parent = frameLoader.load()
      val controller = frameLoader.getController[MapBuilder]()
      
      //val builderLoader = new FXMLLoader(getClass.getResource("ElementBuilder.fxml"))
      //val builder: Parent = builderLoader.load()
      //val builderController = builderLoader.getController[ElementBuilder]()
      
      //controller.pane.children = builder
      //controller.builder = builderController
      
      val scene: Scene = new Scene(root)
      primaryStage.setScene(scene)
      primaryStage.show()
    }
  }
  
  def main(args: Array[String]): Unit = {
    Application.launch(classOf[MBA], args: _*)
  }
}